
require 'treetop'
require_relative 'utils.rb'
require_relative 'types.rb'
require_relative 'stdlib.rb'

DIR = File.dirname(__FILE__)
if File.exists?(DIR + '/DhallishGrammar.rb') and File.mtime(DIR + '/DhallishGrammar.rb') >= File.mtime(DIR + '/DhallishGrammar.treetop')
	require_relative 'DhallishGrammar.rb'
elsif File.writable?(DIR) and does_cmd_exist?('tt') and system("tt '#{DIR + '/DhallishGrammar.treetop'}'")
	require_relative 'DhallishGrammar.rb'
else
	Treetop.load File.join(DIR, 'DhallishGrammar.treetop')
end

module Dhallish

	@@parser = DhallishGrammarParser.new

	def empty_context(basedir=Dir.pwd)
		ctx = { "<#TYPES#>" => Context.new, "<#VALS#>" => Context.new }
		ctx["<#TYPES#>"]["<#DIR#>"] = basedir
		fill_context(ctx["<#VALS#>"], ctx["<#TYPES#>"])
		ctx
	end
	module_function :empty_context

	def evaluate(dhallcode, ctx=nil, expected_type=nil)
		if ctx.nil?; ctx = empty_context() end

		rawast = @@parser.parse dhallcode
		if rawast.nil?
			raise DhallError, "#{@@parser.failure_reason} (line/column: #{@@parser.failure_line}:#{@@parser.failure_column})"
		end

		ast = rawast.to_node

		type = ast.compute_type ctx["<#TYPES#>"]
		res = ast.evaluate ctx["<#VALS#>"]

		if !expected_type.nil?
			if !(type == expected_type)
				raise DhallError, "expression return type missmatch: expected `#{expected_type}`, got: `#{type}`"
			else
				res
			end
		else
			[res, type]
		end
	end
	module_function :evaluate

	def create_ctx(dhallcode, basedir=Dir.pwd)
		empty_ctx = empty_context(basedir)
		ctx, type = evaluate(dhallcode, empty_ctx)
		if type != Types::Record.new({}) or ctx["<#TYPES#>"].nil? or ctx["<#VALS#>"].nil?
			raise DhallError, "return type of dhallcode supplied to `create_ctx` did not return a context (but something of type `#{type}`). use `???`"
		end
		ctx
	end
	module_function :create_ctx

	def create_ctx_from_file(dhallfile, basedir=nil)
		if basedir.nil?; basedir = File.dirname(dhallfile) end
		file = File.open(dhallfile)
		res = create_ctx(file.read, basedir)
		file.close
		res
	end
end
